#include <stdlib.h>
#include "ordonnancement.h"

processus* creer_processus(uint id, uint cout) {
    processus* p = (processus*) malloc(sizeof(processus));
    p->id = id;
    p->cout = cout;
    return p;
}

file* creer_file_vide() {
    file* f = (file*) malloc(sizeof(file));
    f->capacite = 0;
    f->debut = 0;
    f->fin = 0;
    f->charge = 0;
    f->vide = true;
    f->contenu = NULL;
    return f;
}

void detruire_file(file* f) {
    if ( !f->vide )
        for (uint i = f->debut; i < f->fin + f->capacite; ++i)
            free(f->contenu[i % f->capacite]);
    free(f->contenu);
    free(f);
}

processus* retire(file* f) {
    if (f->vide)
        return NULL;
    f->fin = (f->fin + f->capacite - 1) % f->capacite;
    if (f->fin == f->debut)
        f->vide = true;
    f->charge -= f->contenu[f->fin]->cout;
    return f->contenu[f->fin];
}

void ajoute(file* f, processus* p) {
    processus** tmp;
    if ( f->capacite == 0 ) {
        ++f->capacite;
        f->contenu = (processus**) malloc(sizeof(processus*));
    }
    else if ( f->fin == f->debut && !f->vide ) {
        tmp = (processus**) malloc(sizeof(processus*) * f->capacite * 2);
        for (uint i = 0; i < f->fin; ++i)
            tmp[i] = f->contenu[i];
        for (uint i = f->debut; i < f->capacite; ++i)
            tmp[i+f->capacite] = f->contenu[i];
        f->debut += f->capacite;
        f->capacite *= 2;
        free(f->contenu);
        f->contenu = tmp;
    }
    f->contenu[f->fin] = p;
    f->charge += p->cout;
    f->fin = (f->fin + 1) % f->capacite;
    f->vide = false;
}

file* F(uint u0, uint n) {
    uint u = u0;
    file* f = creer_file_vide();
    for (uint i = 0; i <= n; ++i) {
        if (u % 21)
            ajoute(f,creer_processus(u, (u % 10) * ( n % 10 + 1 ) ));
        else
            retire(f);
        u = ( 15091 * u ) % 64007;
    }
    return f;
}

uint taille_file(file* f) {
    return (f->fin + f->capacite) - f->debut;
}

uint milieu_file(file* f) {
    return f->contenu[(((f->fin + f->capacite) - f->debut) / 2) % f->capacite]->id;
}

uint calcul_mu(file* p, uint indice, uint difference_charge, int** memoization) {
    uint cout = p->contenu[(p->fin+p->capacite-indice-1) % p->capacite]->cout;
    uint diff1 = cout + difference_charge;
    uint diff2 = cout < difference_charge ?
        difference_charge - cout : cout - difference_charge;
    int mu_diff1 = memoization[indice+1][diff1] ,
        mu_diff2 = memoization[indice+1][diff2];
    if(mu_diff1 == -1) {
        mu_diff1 = calcul_mu(p, indice+1, diff1, memoization);
    }
    if(mu_diff2 == -1) {
        mu_diff2 = calcul_mu(p, indice+1, diff2, memoization);
    }
    memoization[indice][difference_charge] =
        mu_diff1 < mu_diff2 ? mu_diff1 : mu_diff2;
    return memoization[indice][difference_charge];
}

uint difference_optimale(file* p) {
    if (p->vide)
        return 0;
    uint nb_processus = taille_file(p);
    uint charge_maximale = p->charge;
    int** mu = (int**) malloc(sizeof(int*) * (nb_processus + 1));
    for (uint i = 0; i < nb_processus; ++i) {
        mu[i] = (int*) malloc(sizeof(int) * ( charge_maximale + 1)) ;
        for (uint j = 0; j <= charge_maximale; ++j)
            mu[i][j] = -1;
    }
    mu[nb_processus] = (int*) malloc(sizeof(int) * ( charge_maximale + 1 ));
    for (uint j = 0; j <= charge_maximale; ++j) {
        mu[nb_processus][j] = j;
    }
    return calcul_mu(p,0,0,mu);
}

void equilibrage(file* p, file* a1, file* a2) {
    if (p->vide)
        return;
    uint nb_processus = taille_file(p);
    uint difference_charge = a1->charge < a2->charge ?
        a2->charge - a1->charge : a1->charge - a2->charge;
    uint charge_maximale = p->charge + difference_charge;
    uint cout, diff1, diff2;
    file* f_tmp;
    processus* p_tmp;
    int mu_diff1, mu_diff2;
    int** mu = (int**) malloc(sizeof(int*) * ( nb_processus + 1 ));
    for (uint i = 0; i < nb_processus; ++i) {
        mu[i] = (int*) malloc(sizeof(int) * (charge_maximale+1));
        for (uint j = 0; j <= charge_maximale; ++j)
            mu[i][j] = -1;
    }
    mu[nb_processus] = (int*) malloc(sizeof(int) * ( charge_maximale + 1 ));
    for (uint j = 0; j <= charge_maximale; ++j) {
        mu[nb_processus][j] = j;
    }

    calcul_mu(p,0,difference_charge,mu);

    for (uint indice = 0; !p->vide; ++indice) {
        p_tmp = retire(p);
        cout = p_tmp->cout;
        diff1 = cout + difference_charge;
        diff2 = cout < difference_charge ?
            difference_charge - cout : cout - difference_charge;
        mu_diff1 = mu[indice+1][diff1];
        mu_diff2 = mu[indice+1][diff2];
        if(mu_diff1 < mu_diff2) {
            ajoute(a1,p_tmp);
            difference_charge = diff1;
        }
        else {
            ajoute(a2,p_tmp);
            difference_charge = diff2;
        }
        if(a1->charge < a2->charge) {
            f_tmp = a2;
            a2 = a1;
            a1 = f_tmp;
        }
    }
}
